/***
 * gform - A generic HTML form decoder 
 *
 * 	It takes a HTML form file as parameter and scans
 *	this file for comments beginning with <!--gform ....>
 *	These are taken to be formatting info and delivery commands
 *	for gform.
 *
 *	See gform.txt for description and samples
 *
 *      Swinburne University of Technology and the author will not be 
 *	responsible for any damages or loss suffered or incurred by 
 *	any user or any other person arising in any way from the use
 *	of this software. No warranty of any kind, either express or
 *	implied is given.
 *
 *      Ver 1.01 harry@swin.edu.au Jun95 Swinburne University
 *
 *	Contibutions:
 *		jem@sunsite.unc.edu University of North Caroline   
 *
 *	See CHANGES for change history.
 ***/

#include <stdio.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>

#include "config.h"

#ifdef BSD
#include <sys/malloc.h>
#else
#include <malloc.h>
#endif
#include <stdlib.h>

#define LEN	4096
	
/* keywords for gform */ 
#define DELIVER		"DELIVER"	/* indicates delivery method */
#define MAIL		"MAIL"		/* a delivery method */
#define FFILE		"FILE"		/* a delivery method */
#define PRINT		"PRINT"		/* Send to printer */
#define SUBJECT		"SUBJECT"	/* Subject of mail mesg */
#define REPLY		"REPLY"		/* reply to form submission */
#define GFORM_TAG	"<!GFORM "	/* gform command line keyword */
#define GFORM_TAG2	"<!--GFORM "	/* gform command line keyword */

/* commandline states */
#define NSTATE		0	/* NULL state */ 
#define TSTATE		1	/* text string with variables */
#define DSTATE		2	/* deliver state */
#define RSTATE		3	/* reply form state */

#define SKIP_WHITE(a) 	while (isspace(*a)) a++ 

/* for tildeexpand */
#define ISPUBLIC 1
#define NOTPUBLIC 0

struct entry {
    char *name;
    char *val;
    struct entry *next;
};

/* from cgi-bin/util.c */

char *fmakeword(FILE *f, char stop, int *len);
void unescape_url(char *url);
void plustospace(char *str);
char *tildeexpand(char *filename, int ispublic);

/* func prototypes */

char *decode_commands(char *fname);
char *command(char *line);
void print_header(void);
char *strnupr(char *line, int len);
void error(char *msg, int linecount);
char *insert_value(char *buf, char *vp, int vl);
char *do_escapes(char *line);
int send_to_file(char *fname, char *buf);

struct entry *entries, *head;
int state, linecount;
char *outfile, *address, *subject, DEFSUBJECT[256], *replyfile;

main(int argc, char *argv[])
{
     int len,i, line[LEN];
     char *word, *buf, *tmp, *tmp2, *fname;
     FILE *f1, *f9;

     if (strcmp(getenv("REQUEST_METHOD"),"POST")!=0) 
	error("This program should be referenced with the method=POST\n",0);
     if (strcmp(getenv("CONTENT_TYPE"),
		"application/x-www-form-urlencoded")!=0)
	error("This program can only be used to decode forms\n",0);

      
     len = atoi(getenv("CONTENT_LENGTH"));

     entries = NULL; replyfile=NULL; 
     outfile = NULL; subject=NULL; address=NULL;

     sprintf(DEFSUBJECT,"Form [%s] submission" ,getenv("QUERY_STRING"));
 
/* Grab all the variable names and data from the form */
 
     for(i=0;len && (!feof(stdin));i++) {
        
	if (entries) { 
	    entries->next = (struct entry *) malloc(sizeof(struct entry));
	    entries=entries->next;
	} else { 
	    entries = (struct entry *) malloc(sizeof(struct entry));
	    head = entries;
	}

/* stick all the fields of the HTML Form into the entries struct */

        entries->val = fmakeword(stdin,'&',&len);
        plustospace(entries->val);
        unescape_url(entries->val);
        word = (char *) makeword(entries->val,'=');
        entries->name = strnupr(word, strlen(word));

	entries->next=NULL;

     }

/* We now have all the data from the form, now decode the 
   'GFORM' commands from the HTML form file */

     fname = getenv("QUERY_STRING");

     if (strncmp(WEBHOME, fname, strlen(WEBHOME))!=0) {
	tmp = (char *) malloc(strlen(fname) + strlen(WEBHOME) + 2);
	sprintf(tmp,"%s/%s",WEBHOME,fname);
     }	
     tmp2 = tildeexpand(tmp, NOTPUBLIC);
     buf = decode_commands(tmp2);

/* Now deliver the results */

     if (outfile)	/* Results to a file */ 
		send_to_file(outfile,buf);
     if (address) 	/* or results thru the mail */
		send_email(address,subject,buf);

     print_header();
     if (replyfile) {	/* The HTML file to reply with */

		if (*(replyfile + 1) == '~') {  /*its in a user's public_html dir*/
			tmp2=tildeexpand(replyfile, ISPUBLIC); 
			/*sure its a bit messy, but who cares?*/
		} else {
			tmp2 = (char *) malloc(strlen(WEBHOME)+strlen(replyfile)+1);
			sprintf(tmp2,"%s/%s",WEBHOME,replyfile);
		}
		
		if ((f1 = fopen(tmp2,"r"))==NULL) {
			printf("<P>Failed to open replyfile: %s<P>",
				replyfile);
			exit(1);
		} else 
			while (fread(line,1,LEN,f1)!=0)
				fprintf(stdout,"%s",line);
	        fclose(f1);		
	} else {	/* A default thank you reply */
		printf("<BODY>");
		printf("<TITLE>Thank you</TITLE>");
		printf("<H3>Thank you</H3>");
		printf("</BODY>");
        }

     return(0);
}
/***
 * decode_commands
 *	Read the HTML form file and look for commands 
 *	Replace variables, translate escaped chars
 *	and return the results
 *
 ***/
char *decode_commands(char *fname)
{

     FILE *f1;
     int BRACKET=0,j,escape, vl, variable, endvar=0;
     char line[LEN], *commandline,*start, *end, *trace,
	  *word,
	  msg[64], tmp[256],
	  *buf,		/* The buf of formatted output */ 
	  *vp,
	  *p,
	  ch;

     if ((f1 = fopen(fname,"r")) == NULL) {
	   sprintf(tmp,"File not found: %s",fname); 
	   error(tmp,0);

     }

     linecount=0; 
     buf = (char *) malloc(1); 
     *buf=0;

     while (fgets(line,LEN, f1) !=NULL) {
	linecount++;
	state=NSTATE;

	if (!(commandline = command(line))) continue;

/* fixup any escaped chars */

	commandline = do_escapes(commandline);

	SKIP_WHITE(commandline);

/* A formatting command line */

	if ((state==NSTATE) && (*commandline == '"')) { 
	    state = TSTATE;
	    start = commandline+1;
	    trace=start; end=NULL;

	    variable=0; escape=0;

		
		for (;*trace && !end;trace++) {
		if (*trace=='"') { 
			end = trace;
			*trace=0;
			if (variable) {
				variable=0; start=trace;	
				buf = insert_value(buf,vp,vl);
			} else {
			    buf = (char *) realloc(buf, strlen(buf)+
				(trace-start));
			    strncat(buf, start , (trace-start));
			}
			state=NSTATE;
			if (BRACKET)
				error("Missing bracket",linecount);

		} else if (variable && *trace == ')') { 
			BRACKET--;	 
			endvar=1;	
			vl--;
		} else if (variable) {
			if ((isspace(*trace) || end || endvar) && !BRACKET) {
				variable=0; start=trace;	
				buf = insert_value(buf,vp,vl);
				endvar=0;
			} else
			 	vl++;
		} else if (*trace=='$') { 
			variable=1; vl=0;
			vp = trace+1;
			if (*vp == '(') {
				vp++;
				BRACKET++;
			}
			buf = (char *) realloc(buf, strlen(buf) + 
			      (trace-start));
			strncat(buf, start , (trace-start));
			start = vp;
		}

	     }
	 } 		/* end of formatting command lines */
	 else 		/* command keywords */
	      if (state==NSTATE) { 
		if (strcmp(strnupr(commandline,strlen(DELIVER)),
			    DELIVER)==0) {
			state=DSTATE;
			commandline=commandline+strlen(DELIVER);

			SKIP_WHITE(commandline);

			if (*commandline != '=') 
				error("Missing '='",linecount);
			commandline++;

			SKIP_WHITE(commandline);
	
			if (strcmp(strnupr(commandline,strlen(FFILE)),FFILE)==0) {
				commandline=commandline+strlen(FFILE);

				SKIP_WHITE(commandline);

				if (*commandline !='"')
					error("Expected \"",linecount);
				else {
					p = commandline+1;
					while (*p && *p!='"') p++; 
					*p=0;
					outfile = strdup(commandline+1);
				}
				state=NSTATE;

			} else if (strcmp(strnupr(commandline, strlen(MAIL)),MAIL)==0) {
				commandline=commandline+strlen(MAIL);

				SKIP_WHITE(commandline);
			        
				if (*commandline!='"')
					error("Expected \"",linecount);
				else {
					p=commandline+1;
					while (*p && *p!='"') p++;
					*p=0;
					address=strdup(commandline+1);
					commandline=p+1;
				}
				SKIP_WHITE(commandline);

				if (strcmp(strnupr(commandline,strlen(SUBJECT)),SUBJECT)==0) {
				     commandline=commandline+strlen(SUBJECT);
				     SKIP_WHITE(commandline);

				     if (*commandline!='=')
				     	     error("Expected =",linecount);
				     commandline++;	

				     SKIP_WHITE(commandline);

				     if (*commandline!='"')
					     error("Expected \"",linecount);
				     else {
					     p=commandline+1;
					     while (*p && *p!='"') p++;
					     *p=0;
					     subject=strdup(commandline+1);
				     }	
				} else 
				    subject=DEFSUBJECT;	
			} else {
				sprintf(msg,"Expected %s or %s",FFILE,MAIL); 
				error(msg,linecount);
			}
			state=NSTATE;
		} else if (strcmp(strnupr(commandline,strlen(REPLY)),
			    REPLY)==0) {
			state=RSTATE;
			commandline=commandline+strlen(REPLY);

		        SKIP_WHITE(commandline);
		
			if (*commandline != '=') 
				error("Missing '='",linecount);
			commandline++;

		        SKIP_WHITE(commandline);
			
			if (*commandline != '"') 
				error("Expected \"",linecount);
			else {
				p=commandline+1;
				while (*p && *p!='"') p++;
				*p=0;
				replyfile=strdup(commandline+1);

			}		
			state=NSTATE;
		}
	     }
	} /* end of lines */
	fclose(f1);

/* A bit crude. Handle Escaped '$' signs that have been turned into
   char 1 with do_escapes().
*/ 
	for (j=0;j<strlen(buf);j++)
		if (*(buf+j)==1)
			*(buf+j) = '$';
	return(buf);	    
}
/** 
 * insert_value
 * 	replace a variable with its value
 *
 **/
char *insert_value(char *buf, char *vp, int vl)
{

	char *word, msg[64];

	word = strnupr(vp,vl);
	entries=head;
	while (entries) {
		if (strcmp(word,entries->name)==0) {
		    buf = (char *) realloc(buf,strlen(buf) + strlen(entries->val));
		    strcat(buf,entries->val);
		    break;
		}
		entries=entries->next;
	}
	if (!entries) {
/* Checkboxes don't return values when not selected, so 
   we can't really flag an error if var not found 
  
	    sprintf(msg,"Variable \"%s\" not found",word);
	    error(msg,linecount);
*/
		buf = (char *) realloc(buf, strlen(buf) + 4);
		strcat(buf,"null");
	}
	return(buf);
}

/***
 * do_escapes 
 *	Replaces all escaped chars with actual char its escaped.
 *
 ***/
char *do_escapes(char *line)
{

	char *p, *tmp, *top;

	p = strdup(line);

	if ((tmp = (char *) malloc(2 * strlen(line)))==NULL)
		return(NULL);

	top = tmp;	
	while (*p) {
		if (*p=='\\') {
			p++;
			switch (*p) {
			  case 'n' : *p = 10; break;
			  case 't' : *p = 8; break;
			  case 'r' : *p = 13; break;
			  case 'f' : *p = 12; break;
			  case '$' : *p = 1; break;   /* special case */
			}
		}
		*(tmp++) = *(p++);
	}	
	return(top);
}

/***
 * command
 *	Looks for a 'gform' commandline.
 *	Returns the uppercase commandline (less the leading command tag
 *	or NULL if it is not a commandline.
 *	For backword compatibility, the old GFORM_TAG is still supported
 *	but not really correct.
 *
 ***/
char *command(char *line)
{

	char *commandline, *word, *word2;

	word = strnupr(line,strlen(GFORM_TAG));
	word2 = strnupr(line,strlen(GFORM_TAG2));

	if (strncmp(GFORM_TAG, word,strlen(GFORM_TAG))==0) {
		commandline=line+strlen(GFORM_TAG);
		return(commandline);
	} else if (strncmp(GFORM_TAG2, word2,strlen(GFORM_TAG2))==0) {
		commandline=line+strlen(GFORM_TAG2);
		return(commandline);
	} else
	   return(NULL);
}

/***
 * send_to_file
 *	Send buffer to a file
 *
 ***/
int send_to_file(char *fname, char *buf)
{
	FILE *f;

	if ((f = fopen(fname,"a")) == NULL)
		return(0);

	fprintf(f,"%s",buf);

	fclose(f);
	return(1);
}

/***
 * print_header
 *	Prints the required header for http stuff 
 *
 ***/	
void print_header()
{

	printf("Content-type: text/html\n\n");

	return;
}
/***
 * strnupr
 *	takes a character string and converts 'len' into
 *	uppercase. It then returns a pointer to the uppercase
 *	chars.
 ***/
char *strnupr(char *line, int len)
{
	int i;
	char *newline;

	newline = (char *) malloc(len + 1);

	for (i=0;i<len;i++)
		*(newline+i) = toupper(line[i]);		
	*(newline+len) = '\0';
	
	return(newline);
}

/***
 * error
 *	Prints a simple error message back to the user 
 *
 ***/
void error(char *msg, int linecount)
{
	print_header();
	printf("<P>gform: Error (line %d) %s<P>\n",linecount,msg);
	exit(1);
}
